/*  Copyright (C) 1996  Byron Stanoszek
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 *
 **************************************************************************/

/* snddriv.c */

/* the main program that runs in the background, playing all the music */


#include "sound.h"

int speeds[13]={4410,8820,11025,13230,15435,17640,22050,
                26460,30870,35280,39690,44100,48510};

#define NUMFRAG 12		/* number of fragments in buffer */

int audio=-1;			/* file descriptor of /dev/audio */
int mixer=-1;			/* file descriptor of /dev/mixer */
int sock=-1;			/* file descriptor of /proc/snddriv */
int frag;			/* size of fragments */
int scselect=1;			/* 1 = scale music down at tune change */
int songselect=-1;		/* song to switch to when music scales down */
int fragsremaining=0;		/* for output timer purposes */
unsigned char sndbuf[16384];	/* buffer containing recent sound data */

song_chain *songlist=NULL;	/* list of data songs in the queue */

int nrows=24, ncols=80, interactive=0, scur=0, svol;
unsigned int slvl[NUMFRAG];

#define GRDL_VER "v1.2a2"


void main(argc, argv)
int argc;
char *argv[];
{
  struct timeval timeout;
  song_chain *new, *snext;
  int a, l, opt=0, chap=0, snum=0;
  char *cmd=*argv;

  /* grab any flags on the command line */
  while(argc > 1) {
    --argc; ++argv;
    if(opt) {
      if(opt & 1)
        chap=atoi(*argv);
      if(opt & 2)
        snum=atoi(*argv);
      opt=0;
      continue;
    }
    if(**argv == '-') {
      for(a=1;(*argv)[a];a++) {
        switch((*argv)[a]) {
        case 'c':
          opt|=1;
          break;
        case 'i':
          interactive++;
          break;
        case 's':
          opt|=2;
          break;
        case 'v':
          fprintf(stdout, "%s: GRDL Version Reference %s\n", cmd, GRDL_VER);
          fflush(stdout);
          exit(0);
        default:
          fprintf(stderr, "%s: illegal option -- %c\n", cmd, (*argv)[a]);
          fprintf(stderr, "Usage: %s [-iiv] [-c chapter] [-s song]\n", cmd);
          fflush(stderr);
          exit(1);
        }
      }
      continue;
    }
  }

  if((mixer = open("/dev/mixer", O_WRONLY, 0)) < 0) {
    perror("/dev/mixer");
    fprintf(stderr, "%s: Try recompiling the system kernel with the sound driver configured.\n", cmd);
    exit(1);
  }

  /* set startup volume */
  input_volume(0x4b4b);

  /* ignore system shutdown signals, keep playing :-) */
  signal(SIGTERM, SIG_IGN);

  /* open the input socket */
  if ((sock = open("/proc/snddriv", O_RDONLY|O_NDELAY)) < 0 &&
      interactive != 1) {
    perror("/proc/snddriv");
    fprintf(stderr, "%s: Can't run in non-interactive mode without first compiling the\n", cmd);
    fprintf(stderr, "%s: system kernel with the snddriv patch installed.\n", cmd);
    exit(1);
  }

  if(interactive)
    fix_tty();
  if(interactive != 1) {
    /* close standard input */
    fclose(stdin);
    if(fork() != 0)
      exit(0);
    setpgrp();
  }

  /* load in the startup tune */
  read_song((chap*256)+snum);


 /* main loop */
 while(1) {

  /* update interactive display screen */
  if(interactive)
    update_display(0);

  /* no tune selected yet? */
  if(!songlist->length) {
    timeout.tv_sec=0;
    timeout.tv_usec=100000;
    select(0,0,0,0,&timeout);  /* timer wait 0.1 sec */
    read_input();
    continue;
  }

  /* no audio open yet? */
  if(audio < 0) {
    timeout.tv_sec=0;
    timeout.tv_usec=100000;
    select(0,0,0,0,&timeout);  /* timer wait 0.1 sec */
    set_audio(songlist, 0);
    continue;
  }

  /* check for sound controls */
  if(!songlist->fpause && songselect == -1)
    read_input();

#define SONGEND ((songlist->repeat_mode >= 3 && songlist->repeat > 0) ? \
                 songlist->repeat:songlist->length)

  /* write out more bytes of the song */
  if((l=SONGEND-songlist->c) > frag)
    l = frag;
  songlist->c += (a=read(songlist->fd, &sndbuf, l));
  write(audio, sndbuf, a);

  /* modify clock timer */
  if(--fragsremaining < 0)
    songlist->count+=(double)a/((double)speeds[songlist->speed-18]*
                     (double)songlist->bytesize);

  /* update sound density level */
  if(interactive) {
    if(++scur >= NUMFRAG)
      scur=0;
    svol=slvl[scur];
    slvl[scur]=soundbar();
  }

  /* reach end of tune? */
  if(songlist->c >= SONGEND) {
    /* set data pointer to beginning of song */
    songlist->c = (songlist->repeat_mode>=2)?0:songlist->repeat;
    lseek(songlist->fd, songlist->c+songlist->headsize, SEEK_SET);

    /* check song's speed scaling at repeat */
    if(songlist->scale != -1) {
      if(songlist->scale < songlist->speed)
        songlist->speed--;
      else
        songlist->speed++;
      if(songlist->scale == songlist->speed)
        songlist->scale = -1;
      set_audio(songlist, 1);
      continue;
    }

    /* check number of times left to repeat same song */
    if(songlist->repeat_num > -2)
      --songlist->repeat_num;
    set_status(3);
    if(songlist->fpause) {
      ioctl(audio, SNDCTL_DSP_SYNC);
      if(!(!songlist->repeat_mode || songlist->repeat_mode == 4 ||
         songlist->repeat_num == -1)) {
        read_input();
        continue;
      }
    }

    /* exit song at repeat */
    if(!songlist->repeat_mode || songlist->repeat_mode == 4 ||
       songlist->repeat_num == -1) {
      /* free up song */
      new=songlist->next;
      close(songlist->fd);
      free(songlist);

      songlist=new;
      if(new)
        set_audio(new, 1);
      else
        read_song(0);
      continue;
    }
  }

  /* scale volume down when changing songs */
  if(songselect > -1 && songlist->volume_scale != 0) {
    songlist->volume_scale = 0;
    set_status(7);
  }

  /* tune volume scale up or down */
  if(fragsremaining < 0 && songlist->volume_scale != -1) {
    if(songlist->volume_scale < songlist->volume/1000) {
      songlist->volume-=((songlist->volspeed+1)*10)/(songlist->bytesize);
      if(songlist->volume/1000 <= songlist->volume_scale) {
        songlist->volume=songlist->volume_scale*1000;
        songlist->volume_scale = -1;
        if(songselect > -1) {
          for(new=songlist;new;new=snext) {
            snext=new->next;
            close(new->fd);
            free(new);
          }
          songlist=NULL;
          read_song(songselect);
          songselect=-1;
          continue;
        }
        set_status(7);
      }
      set_status(6);
    } else {
      songlist->volume+=((songlist->volspeed+1)*10)/(songlist->bytesize);
      if(songlist->volume/1000 >= songlist->volume_scale) {
        songlist->volume=songlist->volume_scale*1000;
        songlist->volume_scale = -1;
        set_status(7);
      }
      set_status(6);
    }
    a=(((songlist->volume/1000)+50)*256)+(songlist->volume/1000)+50;
    ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a);
  }
 } /* main loop */
}

void read_song(track)
int track;
{
  song_chain *new;
  wave_header head;
  char soundfile[256];
  int fd=-1, typ;

  /* read new song */
  if(track) {
    sprintf(soundfile,"%s/ch%02x/%02x.wv2", SOUND_LIBRARY, track/256, track%256);
    if((fd = open(soundfile, O_RDONLY, 0)) == -1) {
      sprintf(soundfile, "%s/ch%02x/%02x.wav", SOUND_LIBRARY,
              track/256, track%256);
      if((fd = open(soundfile, O_RDONLY, 0)) == -1)
        typ=2; /* blank input */
      else
        typ=1; /* .wav input */
    } else
      typ=0; /* .wv2 input */
  } else
    typ=2;

  /* open new song pointer */
  new=(song_chain *)malloc(sizeof(song_chain));
  new->next = songlist;
  songlist=new;

  /* read header from file */
  switch(typ) {
  case 2:
    strcpy(new->title, "");
    new->stereo=0;
    new->speed = 24;
    new->sample_size=16;
    new->bytesize = 2;
    new->length = 0;
    new->repeat = 0;
    new->fpause = 0;
    new->repeat_mode = 1;
    new->volume = 50000;
    new->bartime = 0;
    new->barlength = 0;
    break;
  case 1:
    read(fd, &head, sizeof(head));
    new->stereo = (head.modus == 2)?1:0;
    new->speed = speed_to_scale(head.sample_fq);
    new->sample_size = (int)head.bit_p_spl;
    new->bytesize = (int)head.byte_p_spl;
    new->length = head.data_length;
/*  new->repeat = head.data_repeat;
    if(new->repeat > new->length) */
      new->repeat=0;
    new->repeat_mode = 1;
    new->volume = 50000;
    new->bartime = 0;
    new->barlength = 0;
  }

  /* determine song's title, if not given beforehand */
  if(typ != 2) {
    FILE *f;
    char buf[100];

    strcpy(new->title, "\256Generic Wav\257");
    sprintf(soundfile, "%s/songlist", SOUND_LIBRARY);
    if((f=fopen(soundfile, "r"))) {
      sprintf(soundfile, "%04X", track);
      while(fgets(buf, 99, f)) {
        buf[strlen(buf)-1]=0;
        if(!strncmp(buf, soundfile, 4)) {
          strncpy(new->title, buf+5, 22);
          new->title[22]=0;
          break;
        }
      }
      fclose(f);
    }
  }

  new->headsize = (fd == -1)?0:lseek(fd, 0, SEEK_CUR);
  new->default_speed = new->speed;
  new->default_volume = new->volume;
  new->repeat_num = -1;
  new->fpause = 0;
  new->scale = -1;
  new->volspeed = 255;
  new->volume_scale = -1;
  new->scale_toggle = 1;

  new->tune = track;
  new->c = 0;
  new->fd = fd;
  new->count = 0;

  set_audio(new, 0);
}

void set_audio(song, typ)
song_chain *song;
int typ;
{
  static int samplesize, stereo, speed;
  int a;

  if(interactive) {
    for(a=1;a<9;a++)
      set_status(a);
    update_display(1);
  }

  /* turn off music? */
  if(!song->length) {
    if(audio < 0)
      return;
    close(audio);
    audio=-1;

    /* set song volume */
    a=(100*256)+100;
    ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a);
    if(interactive) {
      set_status(9);
      update_display(1);
    }
    return;
  }

  if(audio < 0) {
    /* try to open the audio file */
    if((audio = open("/dev/audio", O_WRONLY, 0)) > -1) {
      a=(NUMFRAG*65536)+10;  /* 2048 bytes per fragment */
      ioctl(audio, SNDCTL_DSP_SETFRAGMENT, &a);
      ioctl(audio, SNDCTL_DSP_GETBLKSIZE, &frag);
      samplesize=-1; stereo=-1; speed=-1;
    } else
      return;
  }

  set_status(9);

  /* set song volume */
  a=(((song->volume/1000)+50)*256)+(song->volume/1000)+50;
  ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a); /* output volume */

  a=(song->sample_size == samplesize && song->stereo == stereo &&
     song->speed == speed)?1:0;
  if(typ && a)
    return;

  /* clear existing sound audio */
  ioctl(audio, typ?SNDCTL_DSP_SYNC:SNDCTL_DSP_RESET);
  fragsremaining=NUMFRAG;
  svol=0;

  if(a)
    return;

  samplesize=song->sample_size;
  ioctl(audio, SNDCTL_DSP_SAMPLESIZE, &song->sample_size);
  stereo=song->stereo;
  ioctl(audio, SNDCTL_DSP_STEREO, &song->stereo);

  /* set new speed */
  speed=song->speed;
  a=speeds[song->speed-18];
  ioctl(audio, SNDCTL_DSP_SPEED, &a);
}

void read_input()
{
  song_chain *new=NULL, *snext;
  static int pc=0, queue=0;
  static unsigned char port[4];
  int a, l, arg=0;

  if(songlist->fpause)
    songlist->fpause=0;

  if(queue >= 0)
    queue--;
  /* process input data */
  while((l=read(sock, port+pc, 1)) == 1) {
    queue=2;

    switch(*port) {
    case WV_PING:
      if(songlist->length)
        set_audio(songlist, 0);
      pc=-1;
      break;
    case WV_SCSELECT_OFF:
      scselect=0;
      pc=-1;
      break;
    case WV_SCSELECT_ON:
      scselect=1;
      pc=-1;
      break;
    case WV_INPUT_OFF:
      input_volume(0);
      pc=-1;
      break;
    case WV_INPUT_ON:
      input_volume(0x4b4b);
      pc=-1;
      break;
    case WV_DEFAULT:
      if(!songlist->length) {
        pc=-1;
        break;
      }
      scselect=1;
      songlist->repeat_mode=1;
      songlist->speed=songlist->default_speed;
      songlist->scale=-1;
      songlist->volume=songlist->default_volume;
      songlist->volume_scale=-1;
      songlist->scale_toggle=1;
      a=(((songlist->volume/1000)+50)*256)+(songlist->volume/1000)+50;
      ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a);
      set_audio(songlist, 0);
      pc=-1;
      break;
    case WV_SPEED:
      if(pc == 1) {
        if(!songlist->length) {
          pc=-1;
          break;
        }
        a=speed_to_scale(*(port+1)*1000);
        songlist->scale=-1;
        if(songlist->speed != a) {
          songlist->speed = a;
          set_audio(songlist, 0);
        } else
          set_status(5);
        pc=-1;
      }
      break;
    case WV_SCALE:
      if(pc == 1) {
        if(!songlist->length) {
          pc=-1;
          break;
        }
        songlist->scale=*(port+1);
        if(songlist->scale < 18)
          songlist->scale = 18;
        if(songlist->scale > 30)
          songlist->scale = 30;
        if(songlist->scale == songlist->speed)
          songlist->scale = -1;
        set_status(5);
        pc=-1;
      }
      break;
    case WV_NORMAL:
      if(!songlist->length) {
        pc=-1;
        break;
      }
      songlist->scale=-1;
      if(songlist->speed != songlist->default_speed) {
        songlist->speed=songlist->default_speed;
        set_audio(songlist, 0);
      } else
        set_status(5);
      pc=-1;
      break;
    case WV_FREE:
      if(songlist->next) {
        new=songlist->next;
        close(songlist->fd);
        free(songlist);
        songlist=new;
        set_audio(new, 0);
      }
      pc=-1;
      break;
    case WV_ALLOC:
    case WV_ALLOCFX:
    case WV_FADEFX:
    case WV_HALFFX:
    case WV_HALFFXSPEED:
    case WV_ALLOCMIX:
      if(pc == ((*port==WV_HALFFXSPEED)?3:(*port==WV_ALLOCMIX)?1:2)) {
        if(songlist->length && (*port==WV_FADEFX || *port==WV_HALFFX ||
           *port==WV_HALFFXSPEED)) {
          if(songlist->volume_scale == -1)
            songlist->volume_scale=songlist->volume/1000;
          songlist->volume=0;
        }
        if(*port == WV_ALLOCMIX)
          arg=*(port+1)+65280;
        else {
          arg=*(port+2);
          arg+=(*(port+1))*256;
        }
        read_song(arg);
        if(!songlist->length && songlist->next) {
          new=songlist->next;
          close(songlist->fd);
          free(songlist);
          songlist=new;
          set_audio(new, 0);
        } else if(*port != WV_ALLOC) {
          songlist->fpause=1;
          songlist->repeat_mode=(*port==WV_HALFFX || *port==WV_HALFFXSPEED)?4:0;
          set_status(4);
          if(pc == 3 && *(port+3) > 17 && *(port+3) < 31 &&
             songlist->speed != *(port+3)) {
            songlist->speed=*(port+3);
            set_audio(songlist, 0);
          }
        }
        pc=-1;
      }
      break;
    case WV_RESTART:
    case WV_MIDWAY:
      if(!songlist->length) {
        pc=-1;
        break;
      }
      ioctl(audio, SNDCTL_DSP_RESET);
      if(*port == WV_RESTART)
        songlist->c=songlist->count=0;
      else {
        songlist->c=songlist->repeat;
        songlist->count=(double)songlist->repeat/
                        ((double)speeds[songlist->speed-18]*
                        (double)songlist->bytesize);
      }
      lseek(songlist->fd, songlist->c+songlist->headsize, SEEK_SET);
      pc=-1;
      break;
    case WV_SELECT:
    case WV_PREV:
    case WV_NEXT:
    case WV_OFF:
      if(*port != WV_SELECT)
        pc=2;
      if(pc != 2)
        break;

      for(new=songlist;new;new=new->next)
        if(!new->next)
          break;

      if(*port == WV_SELECT) {
        arg=*(port+2);
        arg+=(*(port+1))*256;
      } else {
        arg=new->tune;
        if(*port == WV_PREV)
          arg--;
        if(*port == WV_NEXT)
          arg++;
        if(*port == WV_OFF || arg < 0 || arg > 65535)
          arg=0;
      }

      if(arg != new->tune || songselect > -1) {
        if(new->length && new->c && (scselect || songselect > -1))
          songselect=arg;
        else {
          for(new=songlist;new;new=snext) {
            snext=new->next;
            close(new->fd);
            free(new);
          }
          songlist=NULL;
          read_song(arg);
        }
      }
      pc=-1;
      break;
    case WV_REPEATOFF:
      songlist->repeat_mode=0;
      set_status(4);
      pc=-1;
      break;
    case WV_REPEATON:
      songlist->repeat_mode=1;
      set_status(4);
      pc=-1;
      break;
    case WV_REPEATFULL:
      songlist->repeat_mode=2;
      set_status(4);
      pc=-1;
      break;
    case WV_REPEATMIDDLE:
      songlist->repeat_mode=3;
      set_status(4);
      pc=-1;
      break;
    case WV_REPEATEXIT:
      songlist->repeat_mode=4;
      set_status(4);
      pc=-1;
      break;
    case WV_REPEATNUM:
      if(pc == 1) {
        songlist->repeat_num=*(port+1);
        set_status(3);
        pc=-1;
      }
      break;
    case WV_REPEATINF:
      songlist->repeat_num=-1;
      set_status(3);
      pc=-1;
      break;
    case WV_VOLUME:
      if(pc == 1) {
        if(!songlist->length) {
          pc=-1;
          break;
        }
        songlist->volume_scale=-1;
        if(*(port+1) > 50)
          songlist->volume=50000;
        else
          songlist->volume=*(port+1)*1000;
        a=(((songlist->volume/1000)+50)*256)+(songlist->volume/1000)+50;
        ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a);
        set_status(6);
        set_status(7);
        pc=-1;
      }
      break;
    case WV_VOLOFF:
    case WV_VOLDOWN:
    case WV_VOLSOFT:
    case WV_VOLMAX:
      if(!songlist->length) {
        pc=-1;
        break;
      }
      a=*port == WV_VOLOFF?0:*port == WV_VOLDOWN?10:*port == WV_VOLSOFT?38:50;
      if(songlist->scale_toggle) {
        songlist->volume_scale=a;
        set_status(7);
      } else {
        songlist->volume_scale=-1;
        songlist->volume=a*1000;
        a=(((songlist->volume/1000)+50)*256)+(songlist->volume/1000)+50;
        ioctl(mixer, SOUND_MIXER_WRITE_PCM, &a);
        set_status(6);
        set_status(7);
      }
      pc=-1;
      break;
    case WV_VOLSCALE:
      if(pc == 1) {
        if(!songlist->length) {
          pc=-1;
          break;
        }
        songlist->volume_scale=*(port+1);
        if(songlist->volume_scale > 50)
          songlist->volume_scale = 50;
        set_status(7);
        pc=-1;
      }
      break;
    case WV_VOLSPEED:
      if(pc == 1) {
        if(songlist->length) {
          unsigned char xfer=*(port+1);
          songlist->volspeed=xfer;
          set_status(8);
        }
        pc=-1;
      }
      break;
    case WV_SCALEOFF:
      if(songlist->length)
        songlist->scale_toggle=0;
      pc=-1;
      break;
    case WV_SCALEON:
      if(songlist->length)
        songlist->scale_toggle=1;
      pc=-1;
      break;
    case WV_RESET:
      if(songlist->length)
        ioctl(audio, SNDCTL_DSP_RESET);
      pc=-1;
      break;
    case WV_SYNC:
      if(songlist->length)
        ioctl(audio, SNDCTL_DSP_SYNC);
      pc=-1;
      break;
    case WV_WAIT:
      if(pc == 1) {
        sleep(*(port+1));
        pc=-1;
      }
      break;
    case WV_FINISH:
      if(songlist->length)
        songlist->fpause=1;
      pc=-1;
      break;
    default:
      if(songlist->length && *port > 17 && *port < 31) {
        songlist->scale=-1;
        if(songlist->speed != *port) {
          songlist->speed=*port;
          set_audio(songlist, 0);
        } else
          set_status(5);
      }
      pc=-1;
    }
    pc++;
    if(songlist->length && (songlist->fpause || songselect != -1))
      break;
  }

  if(!queue) {
    printf("\e[v"); /* send code to reset /proc/snddriv */
    fflush(stdout);
    lseek(sock, 0L, SEEK_SET);
  }
}

/* change speed in Hz to a number in the scale table speeds[] */
int speed_to_scale(hertz)
int hertz;
{
  int a;

  for(a=0;a<12;a++)
    if(hertz <= speeds[a]+2000)
      return a+18;
  return 30;
}

void fix_tty()
{
  struct termio pty;
  char *ptr;
  int a;

  /* alter lines and columns if available */
#ifdef TIOCGWINSZ
  {
    struct winsize ws;

    if(ioctl(1, TIOCGWINSZ, &ws) != -1) {
      if(ws.ws_row > 22 && ws.ws_row < 251)
        nrows = ws.ws_row;
        if(ws.ws_col > 79 && ws.ws_col < 501)
          ncols = ws.ws_col;
    }
  }
#endif

  ptr=getenv("LINES");
  if(ptr && (a=atoi(ptr)))
    if(a > 22 && a < 251)
      nrows=a;
  ptr=getenv("COLUMNS");
  if(ptr && (a=atoi(ptr)))
    if(a > 79 && a < 501)
      ncols=a;

  if(interactive == 1) {
    /* put terminal into cbreak mode */
    ioctl(0, TCGETA, &pty);
    pty.c_lflag &= ~ICANON;
    pty.c_lflag &= ~ECHO;
    pty.c_iflag &= ~IXON;
    pty.c_iflag &= ~IXOFF;
    ioctl(0, TCSETA, &pty);
    fcntl(0, F_SETFL, O_NONBLOCK);
  }

  /* create screen */
  printf("\e(K\e[m\e[2J\e[H\e[1;31mNetmare Snddriv");
  printf("\e[2;5H\e[34m%s\e[4H\e[33mby Byron Stanoszek\e[m\n", GRDL_VER);
  printf("\e[12H \e[36mElapsed Time-\e[12;25HSong Length-\e[m\n");
  printf("\e[6;32H\e[1mR\e[7;32HL");
  printf("\e[1;22H\e[35m\332\304\304\304\304\304\304\304\304\304\304\304\304\304\304 Status \304\304\304\304\304\304\304\304\304\304\304\304\304\304\277");
  printf("\e[2;22H\263 \e[2;58H \263");
  printf("\e[3;22H\e[35m\300\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\304\331\e[m\n");

  printf("\e[6;9H\e[32mSpeed:\e[7;8HVolume:");
  printf("\e[6;60HTune Number:\e[7;60HCycles Left:\e[8;60HRepeat Mode:");
  printf("\e[9;60HSpeed Scale:\e[10;59HVolume Scale:\e[11;59HVolume Speed:");
  printf("\e[m\e[15;%dr\e[15H", nrows);
  fflush(stdout);
}

void input_volume(amt)
int amt;
{
  int a=amt;

  ioctl(mixer, SOUND_MIXER_WRITE_CD, &a);
  ioctl(mixer, SOUND_MIXER_WRITE_MIC, &a);
  ioctl(mixer, SOUND_MIXER_WRITE_LINE, &a);
#ifdef SOUND_MIXER_WRITE_LINE1
  ioctl(mixer, SOUND_MIXER_WRITE_LINE1, &a);
  ioctl(mixer, SOUND_MIXER_WRITE_LINE2, &a);
  ioctl(mixer, SOUND_MIXER_WRITE_LINE3, &a);
#endif
}

unsigned int soundbar()
{
  int a=0, b=0, c, d;

  if(!songlist->length)
    return 0;

  /* update sound bars */
  if(songlist->stereo) {
    for(c=0;c<64;c++) {
      d=(songlist->bytesize==2)?(128-sndbuf[c*2]):((char)sndbuf[(c*4)+1]);
      if(d > a)
        a=d;
    }
    for(c=0;c<64;c++) {
      d=(songlist->bytesize==2)?(128-sndbuf[(c*2)+1]):((char)sndbuf[(c*4)+3]);
      if(d > b)
        b=d;
    }
  } else {
    for(c=0;c<64;c++) {
      d=(songlist->bytesize==1)?(128-sndbuf[c]):((char)sndbuf[(c*2)+1]);
      if(d > a)
        a=d;
    }
    b=a;
  }

  return (a*256)+b;
}

void update_display(typ)
int typ;
{
  char buf[40];
  int a, b, c;
  double sec;

  /* basic interactive display handles character input */
  if(interactive == 1) {
    if((a=getchar()) != EOF) {
      printf("%d\n", a);
      fflush(stdout);
    }
  }

  if(!typ && !songlist->length)
    return;

  /* update volume density display */
  if(songlist->length) {
    a=svol/256; b=svol%256;
  } else {
    a=0; b=0;
  }
  a*=songlist->volume*17;
  b*=songlist->volume*17;
  a/=50000*128;
  b/=50000*128;
  printf("\e7\e[?25l\e[1;32m\e[6;34H");
  for(c=0;c<16;c++)
    printf("%s%c", c==10?"\e[33m":c==14?"\e[31m":"", b>c?2:1);
  printf("\e[32m\e[7;34H");
  for(c=0;c<16;c++)
    printf("%s%c", c==10?"\e[33m":c==14?"\e[31m":"", a>c?2:1);

  /* count elapsed time */
  printf("\e[36m\e[13H");
  a=speeds[songlist->speed-18]*songlist->bytesize;
  if(!a)
    a=11025;
  sec=songlist->count;
  b=(int)sec;
  sec*=100;
  if(songlist->bartime)
    sprintf(buf, "%5d:%d%4d:%02d:%02d.%02d", 1, 1, b/3600, (b/60)%60,
            b%60, ((int)sec)%100);
  else
    sprintf(buf, "%4d:%02d:%02d.%02d", b/3600, (b/60)%60, b%60,
            ((int)sec)%100);
  printf("%-23.23s\e[m\e8\e[?25h", buf);
  fflush(stdout);
}

void set_status(mode)
int mode;
{
  song_chain *tmp;
  int a, b;
  double sec;

  if(!interactive)
    return;
  printf("\e7\e[1;32m\e[?25l");

  switch(mode) {
  case 1:  /* display speed & song length */
    printf("\e[6;16H");
    a=speeds[songlist->speed-18]*songlist->bytesize;
    if(!a)
      a=11025;
    if(songlist->length)
      printf("%d Hz  ", speeds[songlist->speed-18]);
    else
      printf("         ");
    printf("\e[10;27H\e[37m%04X: %-22.22s\e[13;24H\e[36m", songlist->tune,
           songlist->length?songlist->title:(!songlist->tune?"Sound Off":""));
    sec=((double)songlist->length/(double)a);
    b=(int)sec;
    sec*=100;
    if(songlist->bartime)
      printf("%5d:%d%4d:%02d:%02d.%02d\e[m\e8", 1, 1, b/3600, (b/60)%60,
             b%60, ((int)sec)%100);
    else
      printf("%4d:%02d:%02d.%02d\e[m\e8", b/3600, (b/60)%60, b%60,
             ((int)sec)%100);
    break;
  case 2:  /* display song characteristics */
    printf("\e[1;73H%s", songlist->length?(songlist->stereo?"Stereo":"  Mono"):
           "      ");
    printf("\e[2;73H%s", songlist->length?((songlist->bytesize/
           (songlist->stereo+1) == 2)?"16 Bit":" 8 Bit"):"      ");
    if(songlist->length)
      printf("\e[3;61H%10d Samples", songlist->length/songlist->bytesize);
    else
      printf("\e[3;61H                  ");
    if(songlist->length && songlist->bartime)
      printf("\e[4;69H%3d/%d Time", songlist->bartime,
             songlist->bartime<6?4:songlist->bartime==7?7:
             songlist->bartime==9?9:8);
    else
      printf("\e[4;69H          ");
    if(!songlist->length)
      printf("\e[6;73H      \e[m\e8");
    else {
      for(a=0,tmp=songlist;tmp;tmp=tmp->next,a++);
      printf("\e[6;73H%-6d\e[m\e8", a);
    }
    break;
  case 3:  /* cycles left */
    printf("\e[7;73H");
    if(songlist->repeat_num > -1)
      printf("%-6d\e[m\e8", songlist->repeat_num);
    else
      printf("      \e[m\e8");
    break;
  case 4:  /* repeat mode */
    printf("\e[8;73H%-6.6s\e[m\e8", !songlist->length?"":(
           songlist->repeat_mode==0?"Off":
           songlist->repeat_mode==1?"On":
           songlist->repeat_mode==2?"Full":
           songlist->repeat_mode==3?"Middle":"Exit"));
    break;
  case 5:  /* speed scale */
    if(songlist->scale > 17 && songlist->scale < 31)
      printf("\e[9;73H%-6d\e[m\e8", speeds[songlist->scale-18]);
    else
      printf("\e[9;73H      \e[m\e8");
    break;
  case 6:  /* volume */
    if(songlist->length)
      printf("\e[7;16H%-5d\e[m\e8", songlist->volume);
    else
      printf("\e[7;16H     \e[m\e8");
    break;
  case 7:  /* volume scale */
    if(songlist->volume_scale > -1)
      printf("\e[10;73H%-5d\e[m\e8", songlist->volume_scale*1000);
    else
      printf("\e[10;73H     \e[m\e8");
    break;
  case 8:  /* volume scale speed */
    if(songlist->length)
      printf("\e[11;73H%-3d\e[m\e8", songlist->volspeed);
    else
      printf("\e[11;73H   \e[m\e8");
    break;
  case 9:  /* status bar */
    printf("\e[1;35m\e[2;24H%-34.34s\e[m\e8", !songlist->length?"Sound Off":
           audio == -1?"Waiting on Audio Output...":"Playing Audio");
    break;
  }
}
